from enum import Enum


class ClientType(str, Enum):
    API = "api"
    WIDGET = "widget"
    WEB = "web"
    DASHBOARD = "dashboard"
    DESKTOP = "desktop"
    CHROME_EXTENSION = "chrome_extension"



from typing import Annotated, Optional

from pydantic import BaseModel, Field


class Pagination(BaseModel):
    limit: Annotated[int, Field(ge=0, le=1000)] = 10
    starting_after: Optional[int] = None
    ending_before: Optional[int] = None



import pytest
from pydantic_core import ValidationError

from nuclia_models.worker.proto import Filter, LLMConfig, Operation
from nuclia_models.worker.tasks import (  # type: ignore
    ApplyTo,
    DataAugmentation,
    SemanticModelMigrationParams,
    TaskName,
    TaskStart,
)


def test_task_validation() -> None:
    # Test case 1: DUMMY task accepts None parameters
    _ = TaskStart(name=TaskName.DUMMY, parameters=None)
    with pytest.raises(ValidationError):
        # Test case 2: ASK task doesn't accept None parameters
        _ = TaskStart(name=TaskName.ASK, parameters=None)
    with pytest.raises(ValidationError):
        # Test case 3: SEMANTIC_MODEL_MIGRATOR task doesn't accept None parameters
        _ = TaskStart(name=TaskName.SEMANTIC_MODEL_MIGRATOR, parameters=None)

    with pytest.raises(ValidationError):
        # Test case 4: DUMMY task doesn't accept not None parameters
        _ = TaskStart(
            name=TaskName.DUMMY, parameters=SemanticModelMigrationParams(semantic_model_id="testid")
        )
    with pytest.raises(ValidationError):
        # Test case 5: At least one operation must be defined for ASK task
        _ = TaskStart(
            name=TaskName.ASK,
            parameters=DataAugmentation(
                name="test", on=ApplyTo.FIELD, filter=Filter(), operations=[], llm=LLMConfig()
            ),
        )
    # Test case 6: ASK task with correct parameters
    _ = TaskStart(
        name=TaskName.ASK,
        parameters=DataAugmentation(
            name="test", on=ApplyTo.FIELD, filter=Filter(), operations=[Operation()], llm=LLMConfig()
        ),
    )
    # Test case 7: SEMANTIC_MODEL_MIGRATOR task with correct parameters
    _ = TaskStart(
        name=TaskName.SEMANTIC_MODEL_MIGRATOR,
        parameters=SemanticModelMigrationParams(semantic_model_id="testid"),
    )



import re
from datetime import datetime
from enum import Enum
from typing import Annotated, Any, Generic, Literal, Optional, TypeVar

from pydantic import BaseModel, ConfigDict, EmailStr, Field, field_validator
from pydantic.main import create_model

from nuclia_models.common.client import ClientType
from nuclia_models.common.pagination import Pagination
from nuclia_models.common.user import UserType
from nuclia_models.common.utils import CaseInsensitiveEnum

T = TypeVar("T")


class EventType(CaseInsensitiveEnum):
    # Nucliadb
    VISITED = "visited"
    MODIFIED = "modified"
    DELETED = "deleted"
    NEW = "new"
    SEARCH = "search"
    SUGGEST = "suggest"
    INDEXED = "indexed"
    CHAT = "chat"
    # Tasks
    STARTED = "started"
    STOPPED = "stopped"
    # Processor
    PROCESSED = "processed"


class BaseConfigModel(BaseModel):
    model_config = ConfigDict(extra="forbid")


class GenericFilter(BaseConfigModel, Generic[T]):
    eq: Optional[T] = None
    gt: Optional[T] = None
    ge: Optional[T] = None
    lt: Optional[T] = None
    le: Optional[T] = None
    ne: Optional[T] = None
    isnull: Optional[bool] = None


class StringFilter(GenericFilter[str]):
    like: Optional[str] = None
    ilike: Optional[str] = None


class AuditMetadata(StringFilter):
    key: str


class QueryFiltersCommon(BaseConfigModel):
    id: Optional[BaseConfigModel] = Field(None)
    date: Optional[BaseConfigModel] = Field(None, serialization_alias="event_date")
    user_id: Optional[GenericFilter[str]] = None
    user_type: Optional[GenericFilter[UserType]] = None
    client_type: Optional[GenericFilter[ClientType]] = None
    total_duration: Optional[GenericFilter[float]] = None
    audit_metadata: Optional[list[AuditMetadata]] = Field(
        None, serialization_alias="data.user_request.audit_metadata"
    )
    resource_id: Optional[BaseConfigModel] = None


class QueryFiltersSearch(QueryFiltersCommon):
    question: Optional[StringFilter] = Field(None, serialization_alias="data.user_request.query")
    resources_count: Optional[StringFilter] = Field(
        None, serialization_alias="data.resources_count", json_schema_extra={"cast_to": "int"}
    )
    filter: Optional[BaseConfigModel] = Field(None, serialization_alias="data.request.filter")
    learning_id: Optional[BaseConfigModel] = Field(None, serialization_alias="data.request.learning_id")


class QueryFiltersChat(QueryFiltersSearch):
    rephrased_question: Optional[StringFilter] = Field(
        None, serialization_alias="data.request.rephrased_question"
    )
    answer: Optional[StringFilter] = Field(None, serialization_alias="data.request.answer")
    retrieved_context: Optional[BaseConfigModel] = Field(None, serialization_alias="data.request.context")
    chat_history: Optional[BaseConfigModel] = Field(None, serialization_alias="data.request.chat_context")
    feedback_good: Optional[GenericFilter[bool]] = Field(
        None,
        serialization_alias="data.feedback.good",
        json_schema_extra={"cast_to": "bool"},
        description="True if the feedback provided for the main question is positive.",
    )
    feedback_comment: Optional[StringFilter] = Field(
        None,
        serialization_alias="data.feedback.feedback",
        description="User-provided comment on the feedback for the question.",
    )
    feedback_good_all: Optional[GenericFilter[bool]] = Field(
        None,
        serialization_alias="data.feedback.all",
        json_schema_extra={"cast_to": "bool"},
        description="True if all feedback, including that on the main question and each related text block, is positive.",
    )
    feedback_good_any: Optional[GenericFilter[bool]] = Field(
        None,
        serialization_alias="data.feedback.any",
        json_schema_extra={"cast_to": "bool"},
        description="True if there is any positive feedback on the question itself or any related text block.",
    )
    feedback: Optional[BaseConfigModel] = Field(
        None,
        serialization_alias="data.feedback",
        description="Raw feedback data associated with the question or generative answer, including feedback on related text blocks.",
    )
    model: Optional[StringFilter] = Field(None, serialization_alias="data.request.model")
    rag_strategies_names: Optional[BaseConfigModel] = Field(None, serialization_alias="data.rag_strategies")
    rag_strategies: Optional[BaseConfigModel] = Field(
        None, serialization_alias="data.user_request.rag_strategies"
    )
    status: Optional[GenericFilter[int]] = Field(
        None,
        serialization_alias="data.request.status_code",
        json_schema_extra={"cast_to": "int"},
    )
    time_to_first_char: Optional[BaseConfigModel] = Field(
        None,
        serialization_alias="data.generative_answer_first_chunk_time",
        json_schema_extra={"cast_to": "int"},
    )


def create_dynamic_model(name: str, base_model: type[QueryFiltersChat]) -> type[BaseModel]:
    field_definitions = {}
    field_type_map = {
        "id": int,
        "user_type": Optional[UserType],
        "client_type": Optional[ClientType],
        "total_duration": Optional[float],
        "time_to_first_char": Optional[float],
        "resources_count": Optional[int],
        "feedback_good": Optional[bool],
        "feedback": Optional[dict],
        "status": Optional[int],
        "rag_strategies": Optional[list],
        "rag_strategies_names": Optional[list],
        "chat_history": Optional[list],
        "retrieved_context": Optional[list],
    }
    for field_name in base_model.model_fields.keys():
        field_type = field_type_map.get(field_name, Optional[str])

        field_definitions[field_name] = (field_type, Field(default=None))

    return create_model(name, **field_definitions)  # type: ignore


ActivityLogsQueryResponse = create_dynamic_model(
    name="ActivityLogsQueryResponse", base_model=QueryFiltersChat
)


class ActivityLogsQueryCommon(BaseConfigModel):
    year_month: str

    @field_validator("year_month")
    def validate_year_month(cls, value: str) -> str:
        if not re.match(r"^\d{4}-(0[1-9]|1[0-2])$", value):
            raise ValueError("year_month must be in the format YYYY-MM")
        return value

    @staticmethod
    def _validate_show(show: set[str], model: type[QueryFiltersCommon]) -> set[str]:
        allowed_fields = list(model.model_fields.keys())
        for field in show:
            if field.startswith("audit_metadata."):
                continue
            if field not in allowed_fields:
                raise ValueError(f"{field} is not a field. List of fields: {allowed_fields}")
        return show


SHOW_LITERAL = Literal[tuple(QueryFiltersCommon.model_fields.keys())]  # type: ignore
SHOW_SEARCH_LITERAL = Literal[tuple(QueryFiltersSearch.model_fields.keys())]  # type: ignore
SHOW_CHAT_LITERAL = Literal[tuple(QueryFiltersChat.model_fields.keys())]  # type: ignore
DEFAULT_SHOW_VALUES = {"id", "date"}
DEFAULT_SHOW_SEARCH_VALUES = DEFAULT_SHOW_VALUES | {"question", "resources_count"}
DEFAULT_SHOW_CHAT_VALUES = DEFAULT_SHOW_SEARCH_VALUES | {
    "rephrased_question",
    "answer",
    "rag_strategies_names",
}


class ActivityLogs(ActivityLogsQueryCommon):
    show: set[SHOW_LITERAL] = DEFAULT_SHOW_VALUES  # type: ignore
    filters: QueryFiltersCommon

    @field_validator("show")
    def validate_show(cls, show: set[str]) -> set[str]:
        return cls._validate_show(show=show, model=QueryFiltersCommon)


class ActivityLogsSearch(ActivityLogsQueryCommon):
    show: set[SHOW_SEARCH_LITERAL] = DEFAULT_SHOW_SEARCH_VALUES  # type: ignore
    filters: QueryFiltersSearch

    @field_validator("show")
    def validate_show(cls, show: set[str]) -> set[str]:
        return cls._validate_show(show=show, model=QueryFiltersSearch)


class ActivityLogsChat(ActivityLogsQueryCommon):
    show: set[SHOW_CHAT_LITERAL] = DEFAULT_SHOW_CHAT_VALUES  # type: ignore
    filters: QueryFiltersChat

    @field_validator("show")
    def validate_show(cls, show: set[str]) -> set[str]:
        return cls._validate_show(show=show, model=QueryFiltersChat)


class ActivityLogsSearchQuery(ActivityLogsSearch):
    pagination: Pagination = Pagination()


class ActivityLogsChatQuery(ActivityLogsChat):
    pagination: Pagination = Pagination()


class ActivityLogsQuery(ActivityLogs):
    pagination: Pagination = Pagination()


class DownloadRequestType(str, Enum):
    QUERY = "query"


class DownloadFormat(str, Enum):
    NDJSON = "ndjson"
    CSV = "csv"


class DownloadRequest(BaseModel):
    id: Annotated[int, Field(exclude=True)]
    request_id: str
    download_type: DownloadRequestType
    download_format: DownloadFormat
    event_type: EventType
    requested_at: datetime
    user_id: Annotated[str, Field(exclude=True)]
    kb_id: str
    query: Annotated[dict[Any, Any], Field(exclude=True)]
    download_url: Optional[str]

    # Configuration for Pydantic v2 to handle ORM mapping
    class Config:
        from_attributes = True


class DownloadActivityLogsQueryMixin(BaseModel):
    email_address: Optional[EmailStr] = Field(default=None)
    notify_via_email: bool = Field(default=False)


class DownloadActivityLogsSearchQuery(DownloadActivityLogsQueryMixin, ActivityLogsSearch):
    pass


class DownloadActivityLogsChatQuery(DownloadActivityLogsQueryMixin, ActivityLogsChat):
    pass


class DownloadActivityLogsQuery(DownloadActivityLogsQueryMixin, ActivityLogs):
    pass



from enum import IntEnum
from typing import Optional

from pydantic import BaseModel, Field

# TODO: The models from this file are generated from private protos
# Here we just duplicate the models, but we should migrate the private
# protos to this repository to make them public


class ApplyTo(IntEnum):
    TEXT_BLOCK = 0
    FIELD = 1


class Filter(BaseModel):
    contains: list[str] = Field(default_factory=list)
    resource_type: list[str] = Field(default_factory=list)
    field_types: list[str] = Field(default_factory=list)
    not_field_types: list[str] = Field(default_factory=list)
    rids: list[str] = Field(default_factory=list)
    fields: list[str] = Field(default_factory=list)
    splits: list[str] = Field(default_factory=list)


class KBConfiguration(BaseModel):
    account: str = Field(default="")
    kbid: str = Field(default="")
    onprem: bool = Field(default=False)


class EntityDefinition(BaseModel):
    label: str = Field(description="Entity type")
    description: Optional[str] = Field(default="", description="Description of the entity type")


class EntityExample(BaseModel):
    name: str = Field(description="Name associated with the entity")
    label: str = Field(description="Type of entity")
    example: str = Field(default="", description="Example text where the entity is present")


class RelationExample(BaseModel):
    source: str = Field(description="Entity name from which the relation starts")
    target: str = Field(description="Entity name to which the relation ends")
    label: str = Field(description="Type of relation")
    example: str = Field(default="", description="Example text where the relation is present")


class GraphOperation(BaseModel):
    entity_defs: list[EntityDefinition] = Field(
        default_factory=list, description="Types of entities that will be considered for extraction"
    )
    entity_examples: list[EntityExample] = Field(
        default_factory=list, description="Examples of entities to be sent to the model as few-shot learning"
    )
    relation_examples: list[RelationExample] = Field(
        default_factory=list, description="Examples of relations to be sent to the model as few-shot learning"
    )
    ident: str = Field(default="", title="ID of Operation")


class Label(BaseModel):
    label: str = Field(default="")
    description: str = Field(default="")
    examples: list[str] = Field(default_factory=list)


class LabelOperation(BaseModel):
    labels: list[Label] = Field(default_factory=list)
    ident: str = Field(default="", title="ID of Operation")
    description: str = Field(default="")
    multiple: bool = Field(default=False)


class AskOperation(BaseModel):
    question: str = Field(default="")
    destination: str = Field(default="")
    json: bool = Field(default=False)  # type: ignore


class QAOperation(BaseModel):
    question_generator_prompt: str = Field(default="")
    system_question_generator_prompt: str = Field(default="")
    summary_prompt: str = Field(default="")
    generate_answers_prompt: str = Field(default="")


class ExtractOperation(BaseModel):
    class Model(IntEnum):
        TABLES = 0

    model: "ExtractOperation.Model" = Field(default=0)


class Operation(BaseModel):
    graph: Optional[GraphOperation] = Field(default=None, title="Graph Config")
    label: Optional[LabelOperation] = Field(default=None, title="Label Config")
    ask: Optional[AskOperation] = Field(default=None, title="Ask Config")
    qa: Optional[QAOperation] = Field(default=None, title="QA Config")
    extract: Optional[ExtractOperation] = Field(default=None, title="Extract Config")
    prompt_guard: Optional[bool] = Field(default=False, title="Prompt Guard Config")
    llama_guard: Optional[bool] = Field(default=False, title="Llama Guard Config")


class OpenAIKey(BaseModel):
    key: str = Field(default="")
    org: str = Field(default="")


class AzureOpenAIKey(BaseModel):
    key: str = Field(default="")
    url: str = Field(default="")
    deployment: str = Field(default="")
    model: str = Field(default="")


class HFLLMKey(BaseModel):
    class ModelType(IntEnum):
        LLAMA31 = 0
        QWEN25 = 1

    key: str = Field(default="")
    url: str = Field(default="")
    model: "HFLLMKey.ModelType" = Field(default=0)


class AzureMistralKey(BaseModel):
    key: str = Field(default="")
    url: str = Field(default="")


class PalmKey(BaseModel):
    credentials: str = Field(default="")
    location: str = Field(default="")


class MistralKey(BaseModel):
    key: str = Field(default="")


class AnthropicKey(BaseModel):
    key: str = Field(default="")


class TextGenerationKey(BaseModel):
    model: str = Field(default="")


class HFEmbeddingKey(BaseModel):
    """
    Some models require a specific template (including prefix) to work correctly in each task
    For example Snowflake's Arctic-embed requires a specific prefix to work correctly.
    In that case, the query prompt will be
    ```
    passage_prompt: ""
    query_prompt: "Represent this sentence for searching relevant passages: {}"
    ````
    where {} will be replaced by the actual sentence.
    `passage_prompt` is empty because the model does not require alterations to the sentence to embed is as a passage.
    """

    url: str = Field(default="")
    key: str = Field(default="")
    matryoshka: list[int] = Field(default_factory=list)
    similarity: str = Field(default="")
    size: int = Field(default=0)
    threshold: float = Field(default=0.0)
    passage_prompt: str = Field(default="")
    query_prompt: str = Field(default="")


class UserLearningKeys(BaseModel):
    openai: Optional[OpenAIKey] = Field(default=None)
    azure_openai: Optional[AzureOpenAIKey] = Field(default=None)
    palm: Optional[PalmKey] = Field(default=None)
    anthropic: Optional[AnthropicKey] = Field(default=None)
    claude3: Optional[AnthropicKey] = Field(default=None)
    text_generation: Optional[TextGenerationKey] = Field(default=None)
    mistral: Optional[MistralKey] = Field(default=None)
    azure_mistral: Optional[AzureMistralKey] = Field(default=None)
    hf_llm: Optional[HFLLMKey] = Field(default=None)
    hf_embedding: Optional[HFEmbeddingKey] = Field(default=None)


class OpenAIUserPrompt(BaseModel):
    system: str = Field(default="")
    prompt: str = Field(default="")


class AzureUserPrompt(BaseModel):
    system: str = Field(default="")
    prompt: str = Field(default="")


class HFUserPrompt(BaseModel):
    system: str = Field(default="")
    prompt: str = Field(default="")


class PalmUserPrompt(BaseModel):
    prompt: str = Field(default="")


class AnthropicUserPrompt(BaseModel):
    prompt: str = Field(default="")


class Claude3UserPrompt(BaseModel):
    system: str = Field(default="")
    prompt: str = Field(default="")


class TextGenerationUserPrompt(BaseModel):
    prompt: str = Field(default="")


class MistralUserPrompt(BaseModel):
    prompt: str = Field(default="")


class AzureMistralUserPrompt(BaseModel):
    prompt: str = Field(default="")


class SummaryPrompt(BaseModel):
    prompt: str = Field(default="")


class UserPrompts(BaseModel):
    openai: Optional[OpenAIUserPrompt] = Field(default=None)
    azure_openai: Optional[AzureUserPrompt] = Field(default=None)
    palm: Optional[PalmUserPrompt] = Field(default=None)
    anthropic: Optional[AnthropicUserPrompt] = Field(default=None)
    text_generation: Optional[TextGenerationUserPrompt] = Field(default=None)
    mistral: Optional[MistralUserPrompt] = Field(default=None)
    azure_mistral: Optional[AzureMistralUserPrompt] = Field(default=None)
    claude3: Optional[Claude3UserPrompt] = Field(default=None)


class LLMConfig(BaseModel):
    model: str = Field(default="")
    provider: str = Field(default="")
    keys: Optional[UserLearningKeys] = Field(default=None)
    prompts: Optional[UserPrompts] = Field(default=None)


class DataAugmentation(BaseModel):
    name: str = Field(default="")
    on: ApplyTo = Field(default=0)
    filter: Filter = Field()
    operations: list[Operation] = Field(default_factory=list)
    llm: LLMConfig = Field()
